
#ifdef DEBUG_TRACE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "debug_trace.h"

#if defined(__SYMBIAN32__)
#	if defined (__WINS__)
#		define TRACE_FILE_INSTR "c:\\data\\uae4all\\trace_instraddr.log" 
#		define TRACE_FILE_OPCODE "c:\\data\\uae4all\\trace_opcode.log" 
#		define TRACE_FILE_ADDRESS "c:\\data\\uae4all\\trace_address.log" 
#	else
#		define TRACE_FILE_INSTR "e:\\others\\emulators\\amiga\\trace_instraddr.log" 
#		define TRACE_FILE_OPCODE "e:\\others\\emulators\\amiga\\trace_opcode.log" 
#		define TRACE_FILE_ADDRESS "e:\\others\\emulators\\amiga\\trace_address.log" 
#	endif
#else
#	define TRACE_FILE_INSTR "trace_instraddr.log" 
#	define TRACE_FILE_OPCODE "trace_opcode.log" 
#	define TRACE_FILE_ADDRESS "trace_address.log" 
#endif

static FILE *fTraceInstrAddr = NULL;
static FILE *fTraceOpcode = NULL;
static FILE *fTraceAddress = NULL;

/* Instruction address, opcode, accessed address */
static int traceMode = 0;
/* Put line numbers (access counter) at the beginning of lines. */
bool useCounter = false;
/* Translate B/W/L to b/w/l. Upper case letters are used when data is accesed from 68K engine. */
bool caseSensitive = false;  

static unsigned long counter = 0;

bool TraceStart(int aMode, bool aUseCounter, bool aCaseSensitive) {
	if (NULL == fTraceInstrAddr)
		fTraceInstrAddr = fopen (TRACE_FILE_INSTR, "w");
	else
		return false;

	if (NULL == fTraceOpcode)
		fTraceOpcode = fopen (TRACE_FILE_OPCODE, "w");
	else
		return false;

	if (NULL == fTraceAddress)
		fTraceAddress = fopen (TRACE_FILE_ADDRESS, "w");
	else
		return false;

	traceMode = aMode;
	useCounter = aUseCounter;
	counter = 1;
	
	return true;
}

bool TraceStop() {
	if (NULL != fTraceInstrAddr) {
		fclose (fTraceInstrAddr);
		fTraceInstrAddr = NULL;
	}
	else
		return false;
	
	if (NULL != fTraceOpcode) {
		fclose (fTraceOpcode);
		fTraceOpcode = NULL;
	}
	else
		return false;
	
	if (NULL != fTraceAddress) {
		fclose (fTraceAddress);
		fTraceAddress = NULL;
	}
	else
		return false;
	
	traceMode = 0;
	
	return true;
}

int TraceGetMode() {
	return traceMode;
}

size_t TraceLog (int aModeMask, trace_u32 aInstrAddr, trace_u16 aOpcode, trace_u32 aAddress, const char *aRW, trace_u32 aData, trace_u32 aNewData) {
	char line[80];
	char rw[3] = { aRW[0], aRW[1], '\0' };
	size_t written = 0;
	
	if ((aModeMask & traceMode & TRACE_MODE_INSTR) > 0) {
		if (NULL != fTraceInstrAddr) {
			if (useCounter)
				sprintf (line, "%08u. 0x%08x\n", counter, aInstrAddr);
			else
				sprintf (line, "0x%08x\n", aInstrAddr);
			written += fwrite (line, strlen (line), 1, fTraceInstrAddr);
			if (0 == written)
				return 0;
		}
	}
	
	if ((aModeMask & traceMode & TRACE_MODE_OPCODE) > 0) {
		if (NULL != fTraceOpcode) {
			if (useCounter)
				sprintf (line, "%08u. 0x%04x\n", counter, aOpcode);
			else
				sprintf (line, "0x%04x\n", aOpcode);
			written += fwrite (line, strlen (line), 1, fTraceOpcode);
			if (0 == written)
				return 0;
		}
	}
	
	if ((aModeMask & traceMode & TRACE_MODE_ADDRESS) > 0) {
		if (NULL != fTraceAddress) {
			if (!caseSensitive && (rw[1] < 'a'))
				rw[1] += ('a' - 'A');
			
			switch (rw[1]) {
				case 'b': case 'B':
					if (useCounter)
						sprintf (line, "%08u. %s: 0x%08x (0x%02x)<->(0x%08x)\n", counter, rw, aAddress, aData & 0xFF, aNewData);
					else
						sprintf (line, "%s: 0x%08x (0x%02x)\n", rw, aAddress, aData & 0xFF);
					break;
				case 'w': case 'W':
					if (useCounter)
						sprintf (line, "%08u. %s: 0x%08x (0x%04x)<->(0x%08x)\n", counter, rw, aAddress, aData & 0xFFFF, aNewData);
					else
						sprintf (line, "%s: 0x%08x (0x%04x)\n", rw, aAddress, aData & 0xFFFF);
					break;
				case 'l': case 'L':
					if (useCounter)
						sprintf (line, "%08u. %s: 0x%08x (0x%08x)<->(0x%08x)\n", counter, rw, aAddress, aData /*& 0xFFFFFFFF*/, aNewData);
					else
						sprintf (line, "%s: 0x%08x (0x%08x)\n", rw, aAddress, aData /*& 0xFFFFFFFF*/);
					break;
			}
			written += fwrite (line, strlen (line), 1, fTraceAddress);
			if (0 == written)
				return 0;
		}
	}
	
	if (written > 0)
		counter++;
	
	return written;
}

#endif /* DEBUG_TRACE */
