/*
 * input.cpp
 *
 *  Created on: Aug 9, 2010
 */
#include "input.h"
#include "menu.h"

/* Key/joystick repeat is active ? */
bool gbKeyRepeatOn = false;

extern int mainMenu_keypadConf;

bool CheckKeyJoyRefresh (
		int aLeft, int aRight, int aUp, int aDown,
		int aHit0, int aHit1, int aHit2, int aHit3, int aHit4, int aHit5,
		int aKeyIndex) {
	/* Last joy event */
	static int joyEvent_last = 0;
	/* Last key pressed. SDL index into keyboard snapshot (SDL_GetKeyState()) */
	static int keyEvent_last = 0;
	/* Key ticks */
	static Uint32 ticks_last = 0;
	/* Special delay indicator when keypad is acting as joystick/mouse */
	static bool delay_keypad = false;
	/* Since this function is called only when keys are pressed, total delay
	 * time for special delay (outside of menus) has to be summed up.
	 */
	static Uint32 delay_time = 0;
	Uint32 ticks;
	bool ret_val = true;
	
	int joyEvent = 0;
	
	/* This event */
	if (aUp)
		joyEvent = EKJ_UP;
	else if (aDown)
		joyEvent = EKJ_DOWN;
	else if (aLeft)
		joyEvent = EKJ_LEFT;
	else if (aRight)
		joyEvent = EKJ_RIGHT;
	else if (aHit0)
		joyEvent = EKJ_HIT0;
	else if (aHit1)
		joyEvent = EKJ_HIT1;
	else if (aHit2)
		joyEvent = EKJ_HIT2;
	else if (aHit3)
		joyEvent = EKJ_HIT3;
	else if (aHit4)
		joyEvent = EKJ_HIT4;
	else if (aHit5)
		joyEvent = EKJ_HIT5;
		
	/* Current ticks */
	ticks = SDL_GetTicks();
	
	/* Repeat delay & rate */
	if (((joyEvent > 0) && (joyEvent_last == joyEvent)) ||
			((aKeyIndex > 0) && (keyEvent_last == aKeyIndex))){
		if (!gbKeyRepeatOn) {
			if (delay_keypad) {
				/* Special delay while emulation is running, active only if keypad is set
				 * to act as joystick/mouse - 1s fixed delay for 'clumsy fingers'.
				 */
				/* Summing up of delay intervals, but only if consecutive key presses
				 * are coming at a fast enough rate (key is being held down).
				 */
				if ((ticks - ticks_last) < (1000 / KEY_JOY_REPEAT_RATE)) {
					delay_time += ticks - ticks_last;
					ticks_last = ticks;
				} else {
					delay_time = 0;
					ticks_last = ticks;
					return false;
				}
				if (delay_time < 1000) return false;
				delay_time = 0;
			} else {
				/* Normal key repeat delay */
				if ((ticks - ticks_last) < KEY_JOY_REPEAT_DELAY) return false;
				/* Delay time has passed. Repeat rate is checked from this point on. */
				gbKeyRepeatOn = true;
			}
		} else {
			/* Repeat rate check */
			if (gActiveMenu_last != EMENU_NONE) {
				if ((ticks - ticks_last) < (1000 / KEY_JOY_REPEAT_RATE)) return false;
			} else {
				/* Slower repeat rate outside of menus */
				if ((ticks - ticks_last) < (4 * 1000 / KEY_JOY_REPEAT_RATE)) return false;
			}
		}
	} else {
		/* Special delay outside of menus if keypad acts as joystick/mouse */
		if ((gActiveMenu_last == EMENU_NONE) && (mainMenu_keypadConf > 0) && (aKeyIndex > 0)) {
			/* Start keypad delay and ignore this keystroke */
			delay_keypad = true;
			delay_time = 0;
			ret_val = false;
		} else {
			delay_keypad = false;
			gbKeyRepeatOn = false;
		}
	}
	
	joyEvent_last = joyEvent;
	keyEvent_last = aKeyIndex;
	ticks_last = ticks;
	
	return ret_val;
}
